/***************************************************************************
                          core.c  -  description
                             -------------------
    begin                : Tue Feb 5 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/


#include "core.h"
#include <errno.h>

struct _shmap_lock {
        
        int lock_r;  // read lock
        int lock_w;  // write lock

        key_t key_r; // read  lock key
        key_t key_w; // write lock key
        
        unsigned int readers;
        
        shmap_lock_mode mode;
        };

static int __core_remove_ipcsem ( int semid, int semnum )
{
        return semctl(semid, semnum, IPC_RMID, 0);
}

/*
        returns hash value based on given key
*/
unsigned int _core_get_hashvalue ( const char * key )
{
        int i;
        unsigned int hash = 0;
        unsigned int inc  = (unsigned long) 0x01000000;
        
        
        for ( i = 0; *key && i < 5 ; i++ )
        {
                hash += *key * inc;
                inc = inc >> 8;
                key++;
                
        }
        return hash;
}

key_t __core_keyvalue ( const char * name, char type )
{
        key_t   key;

        if ( name == NULL )
                return IPC_PRIVATE;

        key = ftok(name, type);
        if ( key > 0 ) return key;

        return IPC_PRIVATE;
}

static struct sembuf do_lock[2];
static struct sembuf do_unlock[1];

shmap_lock * _core_lock_init ( void *p_segm, const char * name )
{
        shmap_lock *lock;
        union semun sm_ctl;
        key_t key;

        sm_ctl.val = 1;
        
        lock = sh_mem_malloc(p_segm, sizeof(shmap_lock));

        // read lock
        key = __core_keyvalue(name, 'a');
        lock->key_r  = key;
        lock->lock_r = semget(key, 1, IPC_CREAT|S_IRUSR|S_IWUSR);
        semctl(lock->lock_r, 0, SETVAL, sm_ctl);

        // write lock
        key = __core_keyvalue(name, 'b');
        lock->key_w  = key;
        lock->lock_w = semget(key, 1, IPC_CREAT|S_IRUSR|S_IWUSR);
        semctl(lock->lock_w, 0, SETVAL, sm_ctl);

        do_lock[0].sem_num = 0;
        do_lock[0].sem_op  = -1;
        do_lock[0].sem_flg = SEM_UNDO;
        
        do_unlock[0].sem_num = 0;
        do_unlock[0].sem_op  = 1;
        do_unlock[0].sem_flg = SEM_UNDO;
        
        lock->readers = 0;
        
        return lock;
}

/*
        used to reattach to existing semaphore locks
*/
int _core_lock_attach ( shmap_lock * lock )
{
        do_lock[0].sem_num = 0;
        do_lock[0].sem_op  = -1;
        do_lock[0].sem_flg = SEM_UNDO;

        do_unlock[0].sem_num = 0;
        do_unlock[0].sem_op  = 1;
        do_unlock[0].sem_flg = SEM_UNDO;

        return 1;
}

void _core_lock_dispose ( shmap_lock * lock, void *p_segm )
{
        __core_remove_ipcsem(lock->lock_r, 0);
        __core_remove_ipcsem(lock->lock_w, 0);
        sh_mem_free(p_segm, lock);
}


void _core_lock ( shmap_lock * lock, shmap_lock_mode mode )
{
        if ( mode == SHMAP_LOCK_READ )
        {
                while ( semop(lock->lock_r, do_lock, 1) < 0 );
                lock->readers++;
                if ( lock->readers == 1 )
                        while ( semop(lock->lock_w, do_lock, 1) < 0 );
                while ( semop ( lock->lock_r, do_unlock, 1) < 0 );
        }
        else
        {
                while ( semop(lock->lock_w, do_lock, 1) < 0 );
        }
        
        lock->mode = mode;
}


void _core_unlock ( shmap_lock * lock )
{
        if ( lock->mode == SHMAP_LOCK_READ )
        {
                while ( semop(lock->lock_r, do_lock, 1) < 0 );
                lock->readers--;
                if ( lock->readers == 0 )
                        while ( semop(lock->lock_w, do_unlock, 1) < 0 );
                while ( semop ( lock->lock_r, do_unlock, 1) < 0 );
        }
        else
        {
                while ( semop(lock->lock_w, do_unlock, 1) < 0 );
        }
}


int _core_set_perm ( shmap * map, shmap_perm * perm )
{
        int status;

        // first we change the memory segment permissions
        status = mm_permission(map->mem_segm, perm->mode, perm->uid, perm->gid);

        // if we have a control file, then also change that permission too
        if ( map->opts->p_mem_name != NULL )
        {
                status = chmod(map->opts->p_mem_name, perm->mode);
                status = chown(map->opts->p_mem_name, perm->uid, perm->gid);
        }

        return status;
}

shmap_perm * _core_get_perm ( shmap * map )
{
        int status, mode, uid, gid;
        shmap_perm * ret;

        status = mm_get_permission(map->mem_segm, &mode, &uid, &gid);

        ret = sh_mem_malloc(map->mem_segm, sizeof(shmap_perm));

        if ( ret == NULL )
                return NULL;

        ret->mode = mode;
        ret->uid  = uid;
        ret->gid  = gid;

        return ret;
}
