/***************************************************************************
                          pool.c  -  description
                             -------------------
    begin                : Sun Apr 28 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/


#include "pool.h"

int _sh_pool_add_level ( _shpool * pool )
{
        _shpool_list * newlevel;

        newlevel = sh_mem_malloc(pool->memsegm, sizeof(_shpool_list));

        if ( newlevel == NULL )
                return 0;

        memset(newlevel->ptrlist, 0, sizeof(newlevel->ptrlist));

        newlevel->listsize = POOL_LEVEL_SIZE;
        newlevel->cursor   = 0;
        newlevel->removed  = 0;
        newlevel->next     = NULL;

        if ( pool->current != NULL )
                pool->current->next = newlevel;
        else
                pool->current       = newlevel;

        return 1;
}

/**
 *  Reorganizes the list of pointer in a way, that NULL values will be the last ones.
 */
void _sh_pool_defrag_level ( _shpool_list * level )
{
        int i, j;

        for ( i = 0 ; i < level->listsize ; i++ )
        {
                if ( level->ptrlist[i] == NULL )
                        for ( j = i + 1 ; j < level->listsize ; j++ )
                                if ( level->ptrlist[j] != NULL )
                                {
                                        level->ptrlist[i] = level->ptrlist[j];
                                        level->ptrlist[j] = NULL;
                                        level->removed--;
                                        level->cursor--;
                                        break;
                                }
        }
}

void _sh_pool_add ( _shpool * pool, void * ptr )
{
        // if we are full, then try to defrag it
        if ( pool->current->listsize == pool->current->cursor )
        {
                _sh_pool_defrag_level(pool->current);
        }

        if ( pool->current->listsize == pool->current->cursor )
                pool->current->ptrlist[pool->current->cursor] = ptr;

        if ( pool->current->cursor == pool->current->listsize ) // add new level for pool
        {
                _sh_pool_add_level(pool);
        }
        else
        {
                pool->current->cursor++;
        }
}

/**
 *  Replaces old pointer value in pool with new value
 *  (quite a slow method)
 */
int _sh_pool_replace ( _shpool * pool, void * old, void * new )
{
        int i;
        _shpool_list * level = pool->top;

        while ( level->next != NULL )
        {
                for ( i = 0 ; i < level->listsize ; i++ )
                {
                        if ( level->ptrlist[i] == old )
                        {
                                level->ptrlist[i] = new;
                                return 1;
                        }
                }
        }

        return 0;
}

int _sh_pool_remove ( _shpool * pool, void * ptr )
{
        int i;
        void * old = NULL, * new = NULL;
        _shpool_list * level = pool->top;

        while ( level->next != NULL )
        {
                for ( i = 0 ; i < level->listsize ; i++ )
                {
                        if ( level->ptrlist[i] == old )
                        {
                                level->ptrlist[i] = new;
                                level->removed++;
                                return 1;
                        }
                }
        }

        return 0;
}

_shpool * sh_pool_init ( void * p_mem )
{
        _shpool * pool;

        pool = sh_mem_malloc(p_mem, sizeof(_shpool));

        if ( pool == NULL )
                return NULL;

        pool->memsegm = p_mem;

        _sh_pool_add_level(pool);
        pool->top = pool->current;

        return pool;
}

void sh_pool_freeall ( _shpool * pool )
{
        int i;
        _shpool_list * nextlevel;
        _shpool_list * level = pool->top;

        while ( level != NULL )
        {
                for ( i = 0 ; i < level->listsize ; i++ )
                {
                        if ( level->ptrlist[i] != NULL )
                        {
                                sh_mem_free(pool->memsegm, level->ptrlist[i]);
                                level->ptrlist[i] = NULL;
                        }
                }
                level->cursor = 0;
                level->removed= 0;

                nextlevel = level->next;
                if ( level->next != NULL )
                {
                        sh_mem_free(pool->memsegm, level);
                        pool->top = level->next;
                }
                level = nextlevel;
        }
}

/**
 *  disposes the whole pool, freeing all the resources
 */
void sh_pool_dispose ( _shpool * pool )
{
        void * mem = pool->memsegm;

        // first free all resources
        sh_pool_freeall (pool);

        sh_mem_free(mem, pool->top);
        sh_mem_free(mem, pool);
}

void *sh_pool_malloc ( _shpool *pool, size_t size )
{
        void * ret;

        CHECK_NULL(pool)
        CHECK_NULL(pool->memsegm)

        ret = sh_mem_malloc(pool->memsegm, size);

        _sh_pool_add(pool, ret);

        return ret;
}

void *sh_pool_realloc ( _shpool *pool, void *ptr, size_t size )
{
        void * ret;

        CHECK_NULL(pool)
        CHECK_NULL(pool->memsegm);

        ret = sh_mem_realloc(pool->memsegm, ptr, size);

        if ( ptr != ret )
                _sh_pool_replace(pool, ptr, ret);

        return ret;
}

void sh_pool_free ( _shpool * pool, void *ptr )
{
        CHECK_NULL_(pool)
        CHECK_NULL_(pool->memsegm)

        _sh_pool_remove(pool, ptr);

        sh_mem_free(pool->memsegm, ptr);
}

char *sh_pool_strdup ( _shpool * pool, const char *str )
{
        void * ret;
        CHECK_NULL(pool)
        CHECK_NULL(pool->memsegm)
        CHECK_NULL(str)

        ret = sh_mem_strdup(pool->memsegm, str);

        if ( ret != NULL )
                _sh_pool_add(pool, ret);

        return ret;
}

int sh_pool_lock ( _shpool * pool, mm_lock_mode mode )
{
        CHECK_NULL0(pool)
        CHECK_NULL0(pool->memsegm)

        return sh_mem_lock(pool->memsegm, mode);

}

int sh_pool_unlock ( _shpool * pool )
{
        CHECK_NULL0(pool)
        CHECK_NULL0(pool->memsegm)

        return sh_mem_unlock(pool->memsegm);
}

size_t sh_pool_available ( _shpool * pool )
{
        CHECK_NULL0(pool)
        CHECK_NULL0(pool->memsegm)

        return sh_mem_available(pool->memsegm);
}

int sh_pool_detach ( _shpool * pool )
{
        CHECK_NULL0(pool)
        CHECK_NULL0(pool->memsegm)

        return sh_mem_detach(pool->memsegm);
}

int sh_pool_shmid ( _shpool * pool )
{
        CHECK_NULL0(pool)
        CHECK_NULL0(pool->memsegm)

        return sh_mem_shmid(pool->memsegm);
}

