/***************************************************************************
                          shmap.c  -  description
                             -------------------
    begin                : Tue Feb 5 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/


#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "shmap.h"
#include "shmap_internals.h"

/*
        the default options for a map
*/
shmap_opts __def_opts = {NULL, 1000000, SHMAP_TYPE_DEFAULT, 0, NULL};

/*
        well i don't like the idea, but for now
        we put all the modules into one array so
        upon initialization we go thourgh it all
        over again
*/

shmap_module *__shmap_modules[] = {
        &ll_module,
        &hm_module,
        &btree_module,
        &fifo_module,
        NULL
        };

/*
        allocates a shmap_opts structure with malloc and initializes some variables
*/
shmap_opts * shmap_opts_init ( void )
{
        shmap_opts * ret;

        ret = malloc(sizeof(shmap_opts));

        ret->flags = 0;
        ret->p_mem = NULL;
        ret->p_mem_name = NULL;
        ret->size  = 1000000;
        ret->type  = SHMAP_TYPE_DEFAULT;

        return ret;
}

/*
        sets the memory segment for optionset
*/
void shmap_opts_setmem ( shmap_opts * p_opts, void * p_mem )
{
        CHECK_NULL_(p_opts)

        p_opts->p_mem =p_mem;
}

/*
        sets the map type to new value
*/
void shmap_opts_settype ( shmap_opts * p_opts, char map_type )
{
        CHECK_NULL_(p_opts)

        p_opts->type = map_type;
}

void shmap_opts_setsize ( shmap_opts * p_opts, size_t size )
{
        CHECK_NULL_(p_opts)

        p_opts->size = size;
}

void shmap_opts_setflag ( shmap_opts * p_opts, int option )
{
        CHECK_NULL_(p_opts)

        p_opts->flags |= option;
}

void shmap_opts_unsetflag ( shmap_opts * p_opts, int option )
{
        CHECK_NULL_(p_opts)

        p_opts->flags ^= option;
}

/*
        sets the memory segment name

        NOTE: this will have to be presented as file name with FULL PATH
        it can also be NULL, in which case the memory segment can not be
        attached to by external programs
*/
void shmap_opts_setname ( shmap_opts * p_opts, char * filename )
{
        CHECK_NULL_(p_opts)

        p_opts->p_mem_name = filename;
}

/*
        extract directory name from the whole file path

        NOTE: you'll need to specify input buffers and buflens yourself
*/
int __shmap_dirname ( const char * path , char * dirname, int dirlen, char * filename, int filelen )
{
        char * tmp, * ret;

        if ( path == NULL )
        {
                memset(dirname, 0, dirlen);
                memset(filename, 0, filelen);
                return 0;
        }

        ret = strdup(path);

        tmp   = strrchr(ret, '/');
        * tmp = '\0';
        tmp++;

        if ( strlen(ret) == 0 )
                strncpy(dirname, "/", dirlen);
        else
                strncpy(dirname, ret, dirlen);

        strncpy(filename, tmp, filelen);

        free(ret);

        return 1;
}

/*
        this function will try to attach to an existing hashmap
        if no map exists, then NULL is returned or the pointer to
        existing hashmap. also the semaphore locks will be attached

        dirpath is the directory name where the descriptor file exists
        map_file_name is the full path to description file
          yes .. there is duplicate information
*/
shmap * _shmap_attach ( char * dirpath,  char * map_file_name )
{
        int i, shmid, memid;
        FILE * fd;
        char p_shmap[15], p_memory[15], p_shmid[15];
        struct stat buf;
        shmap * ret = NULL;

        CHECK_NULL(map_file_name)

        if ( stat(map_file_name, &buf) == -1 )
                return NULL;

        if ( (fd = fopen(map_file_name, "rt")) == NULL )
                return NULL;

        fgets(p_shmap,  sizeof(p_shmap),  fd);
        fgets(p_memory, sizeof(p_memory), fd);
        fgets(p_shmid,  sizeof(p_shmid),  fd);

        fclose(fd);

        for ( i = 0 ; i < sizeof(p_shmap) ; i++ )
                if ( !isdigit(p_shmap[i]) )
                        p_shmap[i] = '\0';

        for ( i = 0 ; i < sizeof(p_memory) ; i++ )
                if ( !isdigit(p_memory[i]) )
                        p_memory[i] = '\0';

        for ( i = 0 ; i < sizeof(p_shmid) ; i++ )
                if ( !isdigit(p_shmid[i]) )
                        p_shmid[i] = '\0';

        ret = (shmap *) atoi(p_shmap);
        memid = atoi(p_memory);
        shmid = atoi(p_shmid);

        if ( ret == NULL )
                return NULL;

        if ( memid != (int) ret->mem_segm || shmid != sh_mem_shmid(ret->mem_segm) )
                return NULL;

        // now reconnect the locks
        // what we should do here, add better error handling
        _core_lock_attach(ret->locks);

        return ret;
}

/*
        initializes new hashmap structure
*/
shmap * shmap_init ( shmap_opts * p_opts )
{
        int i;
        void  * mem;
        shmap * map;
        FILE  * fd = NULL;
        char    pathname[256], filename[256];
        shmap_perm perm;
        struct stat stat;

        if ( p_opts == NULL )
                p_opts = SHMAP_DEFAULT;

        __shmap_dirname ( p_opts->p_mem_name, pathname, sizeof(pathname), filename, sizeof(filename));

        // initialize the memory
        if ( p_opts->p_mem == NULL )
                mem = sh_mem_init(p_opts->size, pathname);
        else
                mem = p_opts->p_mem;

        // only try to attach to existing external map if supply NULL value for memory segment
        if ( p_opts->p_mem == NULL && (map = _shmap_attach(pathname, p_opts->p_mem_name)) != NULL )
        {
                map->attached++;
                return map;
        }

        // make the map file if we need to
        // this is used to connect to this map by external programs
        if ( p_opts->p_mem_name != NULL )
        {
                if ( (fd = fopen(p_opts->p_mem_name, "wt")) == NULL )
                {
                        return NULL; // we could not open the required file that we needed to
                }
        }

        if ( mem == NULL )
        {
                return NULL;
        }

        // allocate the memory
        map = sh_mem_malloc(mem, sizeof(shmap));
        
        if ( map == NULL ) // malloc failed and therefore we exit with error
        {
                if ( fd != NULL ) // also release the file that we opened
                {
                        unlink(p_opts->p_mem_name);
                        fclose(fd);
                }
                return NULL;
        }
        map->attached = 1;
        
        if ( p_opts->type == 0 )
                p_opts->type = SHMAP_TYPE_DEFAULT;
        
        // search the placeholder for correct type of functions
        map->func_pos = 0;
        for ( i = 0 ; __shmap_modules[i] != NULL ; i++ )
        {
                if ( __shmap_modules[i]->module_type == p_opts->type )
                {
                        map->func_pos = i;
                        break;
                }
        }
        
        // initialize the toplevel
        map->toplevel = __shmap_modules[map->func_pos]->shmap_init(mem, p_opts);

        if ( __shmap_modules[map->func_pos]->shmap_cache_init != NULL )
                map->cache = __shmap_modules[map->func_pos]->shmap_cache_init(mem);
        map->mem_segm = mem;

        // initialize locks
        map->locks = _core_lock_init(mem, pathname);
        // initialize item count
        map->items = 0;
        // also increase this memory segments map count
        sh_mem_addmap(map->mem_segm);

        if ( fd )
        {
                //fprintf(fd, "%d\n%d\n%d", map, map->mem_segm, sh_mem_shmid(map->mem_segm));
                fclose(fd);
        }

        map->opts = sh_mem_malloc(map->mem_segm, sizeof(shmap_opts));

        if ( map->opts != NULL )
        {
                map->opts->flags = p_opts->flags;
                map->opts->p_mem = p_opts->p_mem;
                map->opts->p_mem_name = sh_mem_strdup(map->mem_segm, p_opts->p_mem_name);
                map->opts->size  = p_opts->size;
                map->opts->type  = p_opts->type;
        }
        else
        {
                sh_mem_free(map->mem_segm, map->opts->p_mem_name);
                sh_mem_free(map->mem_segm, map->opts);
                sh_mem_free(map->mem_segm, map);
                return NULL;
        }

        // if we created a file, then we would like to know its permissions, so
        //   that we can set these permissions on shared memory segment as well
        if ( map->opts->p_mem_name != NULL )
        {
                if ( lstat(map->opts->p_mem_name, &stat) )
                {
                        perm.uid  = stat.st_uid;
                        perm.gid  = stat.st_gid;
                        perm.mode = stat.st_mode;
                        _core_set_perm(map, &perm);
                }
        }

        return map;
}

void * shmap_get_value ( shmap * map, const char * key, int cached )
{
        _shmap_value * val;

        CHECK_NULL(map)

        CORE_LOCK(map->locks, SHMAP_LOCK_READ);

        if ( __shmap_modules[map->func_pos]->shmap_get_value != NULL )
                val = __shmap_modules[map->func_pos]->shmap_get_value (map, key, cached);
        else
        {
                CORE_UNLOCK(map->locks);
                return NULL;
        }

        CORE_UNLOCK(map->locks);
        if ( val == NULL )
                return NULL;
        else
                return val->value;
}

/*
        add a string to hashmap
        key and value both will be string duplicated to shared memory segment
*/
int shmap_add_value ( shmap * map, const char * p_key, const char * p_value )
{
        int status;
        char * value, * key;

        CHECK_NULL0(map)

        if ( __shmap_modules[map->func_pos]->shmap_add_value != NULL )
        {
                value = sh_mem_strdup(map->mem_segm, p_value);
                if ( value == NULL )
                        return 0;

                key = sh_mem_strdup(map->mem_segm, p_key);
                if ( key == NULL )
                        return 0;

                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_add_value(map, key, value, __SHMAP_FREE_VALUE|__SHMAP_FREE_KEY, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        if ( status == 2 ) // new items was added, not replaced
                        {
                                map->items++;
                        }
                        return 1;
                }
                else
                {
                        return 0;
                }
        }
        else
                return 0;
}


/*
        only add pointer to map, do not copy the content
        key will be duplicated to shared memory segment, value will not be duplicated
*/
int shmap_add_value_p ( shmap * map, const char * p_key, void * p_value )
{
        int status;
        char * key;

        CHECK_NULL0(map)

        if ( __shmap_modules[map->func_pos]->shmap_add_value != NULL )
        {
                key = sh_mem_strdup(map->mem_segm, p_key);
                if ( key == NULL )
                        return 0;

                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_add_value(map, key, p_value, __SHMAP_FREE_KEY, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        if ( status == 2 ) // new items was added, not replaced
                                map->items++;
                        return 1;
                }
                else
                {
                        return 0;
                }
        }
        else
                return 0;
}

/*
        binary safe adding to map
        key and value will be duplicated to shared memory segment
*/
int shmap_add_value_b ( shmap * map, const char * p_key, void * p_value, size_t value_size )
{
        int status;
        char * key, * value;

        CHECK_NULL0(map)

        if ( __shmap_modules[map->func_pos]->shmap_add_value != NULL )
        {
                key = sh_mem_strdup(map->mem_segm, p_key);
                if ( key == NULL )
                        return 0;

                value = sh_mem_malloc(map->mem_segm, value_size);
                if ( value == NULL )
                        return 0;
                memcpy(value, p_value, value_size);

                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_add_value(map, key, value, 0, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        if ( status == 2 ) // new items was added, not replaced
                                map->items++;
                        return 1;
                }
                else
                {
                        return 0;
                }
        }
        else
                return 0;
}

int shmap_del_value ( shmap * map, const char * key )
{
        int status;
        CHECK_NULL0(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_del_value != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_del_value(map, key, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        map->items--;
                        return 1;
                }
                else
                {
                        return 0;
                }
        }
        else
                return 0;
}

char * shmap_get_sub ( shmap * map, const char * key, int keylen, int pos, int cached )
{
        _shmap_value * val;
        CHECK_NULL(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_get_sub != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_READ);
                val = __shmap_modules[map->func_pos]->shmap_get_sub(map, key, keylen, pos, cached);
                CORE_UNLOCK(map->locks);
                if ( val != NULL )
                        return (char *) val->key;
                else
                        return NULL;
        }
        else
                return 0;

        return 0;
}

int shmap_del_sub ( shmap * map, const char * key, int keylen )
{
        int status;
        CHECK_NULL0(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_del_sub != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_del_sub(map, key, keylen, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        map->items -= status;
                        return 1;
                }
                else
                {
                        return 0;
                }
        }
        else
                return 0;
}

int shmap_clear ( shmap * map )
{
        int status;
        CHECK_NULL0(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_clear != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_clear(map, NULL);
                CORE_UNLOCK(map->locks);

                if ( status )
                {
                        map->items = 0;
                        return 1;
                }
                else
                        return 0;
        }
        else
                return 0;
}


/*

        we will destroy the memory segment as well, after that noone can attach to this memory
        segment

*/
int shmap_dispose ( shmap * map )
{
        int status;
        void * mem;
        shmap_lock * locks;

        CHECK_NULL0(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_dispose != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_WRITE);
                status = __shmap_modules[map->func_pos]->shmap_dispose(map, NULL);

                if ( status )
                {
                        mem = map->mem_segm;
                        locks = map->locks;


                        if ( map->opts && map->opts->p_mem_name )
                                unlink(map->opts->p_mem_name);
                        // only free the resource if we are only one using it
                        // otherwise reduce the attached count for this map
                        if ( map->attached == 1 )
                                sh_mem_free(mem, map);
                        else
                                map->attached--;
                        CORE_UNLOCK(locks);
                        _core_lock_dispose(locks, mem);
                        // we mark this segment for deletion only if
                        // this map is the only on on it but we don't care
                        // about attached count to this map, because the segment
                        // will be deleted after the last process has detached
                        if ( sh_mem_mapcount(mem) == 1 )
                                sh_mem_dispose(mem);
                        else
                                sh_mem_remmap(mem);
                        return 1;
                }
                else
                {
                        CORE_UNLOCK(map->locks);
                        return 0;
                }
        }
        else
        {
                CORE_UNLOCK(map->locks);
                return 0;
        }
                
}

/*

        Frees a resource allocated by shmap library

        Used for example with SHMAP_TYPE_FIFO and shmap_get.
        See doc page for

 */
void shmap_free_resource ( shmap * map, void * resource )
{
        CHECK_NULL_(map)
        CHECK_NULL_(map->mem_segm)

        sh_mem_free(map->mem_segm, resource);
}

/*
        detaches from the memory segment that map uses
*/
int shmap_detach ( shmap * map )
{
        CHECK_NULL0(map)
        CHECK_NULL0(map->mem_segm)

        map->attached--;
        return sh_mem_detach(map->mem_segm);
}

void shmap_empty_cache ( shmap * map )
{
        CHECK_NULL_(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_cache_clean != NULL )
        {
                __shmap_modules[map->func_pos]->shmap_cache_clean(map->cache);
        }
}

void shmap_dump ( shmap * map, int with_values )
{
        CHECK_NULL_(map)
        
        if ( __shmap_modules[map->func_pos]->shmap_dump != NULL )
        {
                CORE_LOCK(map->locks, SHMAP_LOCK_READ);
                __shmap_modules[map->func_pos]->shmap_dump(map, with_values);
                CORE_UNLOCK(map->locks);
        }
}

/*
        returns the memory allocation structure pointer
*/
void * shmap_memsegm ( shmap * map )
{
        CHECK_NULL(map)

        return map->mem_segm;
}

/*
        returns the amount of items in hashmap
*/
unsigned int shmap_items ( shmap * map )
{
        CHECK_NULL0(map)

        return map->items;
}

/*
        returns the amount of bytes available
          in this hashmap
*/
size_t shmap_available ( shmap * map )
{
        CHECK_NULL0(map)

        if ( map->mem_segm != NULL )
                return sh_mem_available(map->mem_segm);
        else
                return 0;
}


/*******************************

        lock based functions

        WARNING BE VERY CAREFUL WHEN YOU USE THEM
        YOU MIGHT END UP IN DEADLOCK SITUATIONS

********************************/

/*
        initialize a brand new lock

        returns the handle to newly created lock
*/
shmap_lock * shmap_lock_init ( void * p_mem, const char * path )
{
        void * ret;
        char pathname[256], filename[256];
        CHECK_NULL(p_mem)

        __shmap_dirname((char *)path, pathname, sizeof(pathname), filename, sizeof(filename));
        ret = _core_lock_init(p_mem, pathname);

        return ret;
}


/*
        returns the lock handle of existing map

        of course NULL is returned, if the map is NULL or
        the map lock is NULL
*/
shmap_lock * shmap_lock_get ( shmap * map )
{
        CHECK_NULL(map)

        return map->locks;
}

/*
        set a new lock to a map
        NOTES: * it does not release the old map, you need to take
                 care of it yourself.
               * if lock is NULL, then the lock is not replaced
                 if you want to remove a lock, then use shmap_lock_dispose
*/
int shmap_lock_set ( shmap * map, shmap_lock * lock )
{
        CHECK_NULL0(map)
        CHECK_NULL0(lock)

        map->locks = lock;

        return 1;
}

/*
        deletes a map lock but does not free any resources
        use shmap_lock_dispose for that
*/
int shmap_lock_del ( shmap * map )
{
        CHECK_NULL0(map)

        map->locks = NULL;
        return 1;
}

/*
        frees the resources help by lock
        NOTE: you need to supply the correct memory segment, otherwise
                you can end up in a segfault
*/
int shmap_lock_dispose ( void * p_mem, shmap_lock * lock )
{
        CHECK_NULL0(lock)
        CHECK_NULL0(p_mem)

        _core_lock_dispose(lock, p_mem);

        return 1;
}

int shmap_dolock ( shmap_lock * lock, shmap_lock_mode mode )
{
        CHECK_NULL0(lock)

        _core_lock(lock, mode);

        return 1;
}


int shmap_unlock ( shmap_lock * lock )
{
        CHECK_NULL0(lock)

        _core_unlock(lock);

        return 1;
}


/*
        Get map permissions
 */
shmap_perm * shmap_get_permissions ( shmap * map )
{
        return _core_get_perm(map);
}

/*
        Set map permissions
 */
int shmap_set_permissions ( shmap * map, shmap_perm * perm )
{
        return _core_set_perm(map, perm);
}
