/***************************************************************************
                          shmap_btree.c  -  description
                             -------------------
    begin                : Sat May 11 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/

#include "shmap_btree.h"

shmap_module btree_module =
        {
        SHMAP_BTREE_MODULE,
        SHMAP_BTREE_VERSION,
        SHMAP_TYPE_BTREE,
        shmap_btree_init,
        shmap_btree_get,
        shmap_btree_add,
        shmap_btree_del,
        shmap_btree_get_sub,
        shmap_btree_del_sub,
        shmap_btree_clear,
        shmap_btree_dispose,
        NULL,
        NULL,
        NULL,
        NULL
        };

void * shmap_btree_init ( void * mem_segm, void * p_opts )
{
        _shmap_btree_top * top;

        top = sh_mem_malloc(mem_segm, sizeof(_shmap_btree_top));
        top->left_items  = 0;
        top->right_items = 0;
        top->tree        = NULL;

        return top;
}

/**
 *
 *   Returns value item from btree tree
 *
 *   cached - ignored
 */
_shmap_value * shmap_btree_get ( shmap * map, const char * key, int cached )
{
        _shmap_btree * btree;

        CHECK_NULL(key)

        btree = _shmap_btree_get(map->toplevel, key, strlen(key), _core_get_hashvalue(key), 0);

        if ( btree ) return btree->value;
        else return NULL;
}

int shmap_btree_add ( shmap * map, const char * key, void * value, int opts, shmap_cleanup cleanup )
{
        int keyhash;
        _shmap_value * value_item;

        keyhash = _core_get_hashvalue(key);

        value_item = shmap_value_init(map->mem_segm, (void *)key, keyhash, value, opts);

        return _shmap_btree_add(map->mem_segm, map->toplevel, value_item, cleanup);
}

int shmap_btree_del ( shmap * map, const char * key, shmap_cleanup cleanup )
{
        _shmap_btree * btree;

        CHECK_NULL0(key)

        btree = _shmap_btree_get(map->toplevel, key, strlen(key), _core_get_hashvalue(key), 0);

        if ( btree != NULL )
                _shmap_btree_remove(map->mem_segm, map->toplevel, btree, cleanup);

        return 1;
}

_shmap_value * shmap_btree_get_sub ( shmap * map, const char * key, int keylen, int pos, int cached )
{
        _shmap_btree * btree;

        btree = _shmap_btree_get(map->toplevel, key, keylen, _core_get_hashvalue(key), pos);

        CHECK_NULL(btree)

        return btree->value;
}

int shmap_btree_del_sub ( shmap * map, const char * key, int keylen, shmap_cleanup cleanup )
{
        int keyhash;
        _shmap_btree * btree;

        keyhash = _core_get_hashvalue(key);

        while ( (btree = _shmap_btree_get(map->toplevel, key, keylen, keyhash, 0)) )
        {
                _shmap_btree_remove(map->mem_segm, map->toplevel, btree, cleanup);
        }

        return 1;
}

int shmap_btree_clear ( shmap * map, shmap_cleanup cleanup )
{
        return _shmap_btree_clear(map->mem_segm, map->toplevel, cleanup);
}

int shmap_btree_dispose ( shmap * map, shmap_cleanup cleanup )
{
        return _shmap_btree_clear(map->mem_segm, map->toplevel, cleanup);
}


/**************************************************
 *
 *   btree internal structure functions
 *
 **************************************************/

_shmap_btree * _shmap_btree_get ( _shmap_btree_top * toplevel, const char * key, int keylen, int keyhash, int pos )
{
        int cpos = 0;
        _shmap_btree * item = toplevel->tree;

        while ( item )
        {
                if ( item->value && item->value->key && keyhash == item->value->keyhash && !strncmp(key, item->value->key, keylen) )
                {
                        if ( pos == cpos ) // if this is the one we are looking for
                                return item;
                        else
                                cpos++;
                }

                // navigation between left and right
                if ( keyhash <= item->value->keyhash )
                        item = item->left;
                else
                if ( keyhash > item->value->keyhash )
                        item = item->right;
        }

        return NULL;
}


int _shmap_btree_remove ( void * p_mem, _shmap_btree_top * toplevel, _shmap_btree * btree, shmap_cleanup cleanup )
{
        _shmap_btree * tmp;
        int left;
        // first we unplug it
        // if we are NOT dealing with toplevel item
        if ( btree->parent )
        {
                if ( toplevel->tree->value->keyhash < btree->value->keyhash ) // left item
                        left = 1;
                else // right item
                        left = 0;

                if ( btree->parent->left == btree )
                {
                        // we connect left tree to parents left tree
                        btree->parent->left = btree->left;
                        _shmap_btree_item_place(p_mem, toplevel, btree->right, cleanup);
                }
                else
                {
                        // we connect the right tree to parents right tree
                        btree->parent->right = btree->right;
                        _shmap_btree_item_place(p_mem, toplevel, btree->left, cleanup);
                }

                // and now set the balance count on both sides of the tree
                left ? toplevel->left_items-- : toplevel->right_items--;
                _shmap_btree_item_delete(p_mem, btree, cleanup);
        }
        else
        if ( !btree->parent ) // if we are dealing with toplevel item
        {
                tmp = toplevel->tree;
                // more on the left side
                if ( toplevel->left_items - toplevel->right_items > 0 )
                {
                        toplevel->tree = toplevel->tree->left;
                        _shmap_btree_item_place(p_mem, toplevel, tmp->right, cleanup);
                }
                else // more on the right side
                {
                        toplevel->tree = toplevel->tree->right;
                        _shmap_btree_item_place(p_mem, toplevel, tmp->left, cleanup);
                }
        }

        return 1;
}


int _shmap_btree_add ( void * p_mem, _shmap_btree_top * top, _shmap_value * value, shmap_cleanup cleanup )
{
        _shmap_btree * item;

        item = _shmap_btree_item_init(p_mem, value);

        return _shmap_btree_item_place(p_mem, top, item, cleanup);
}


/**
 *
 *  Parses though all the items in the btree and deletes them
 *
 *
 */
int _shmap_btree_clear ( void * p_mem, _shmap_btree_top * toplevel, shmap_cleanup cleanup )
{
        _shmap_btree * tmp;
        _shmap_btree * btree = toplevel->tree;

        toplevel->tree = NULL;

        while ( btree )
        {
                // if this is some kind of edge level element, then remove it
                if ( btree->left == NULL && btree->right == NULL )
                {
                        tmp = btree;
                        btree = btree->parent;

                        // unlink parent if there is a parent to unlink
                        if ( btree )
                        {
                                if ( btree->left == tmp )
                                        btree->left = NULL;
                                else
                                        btree->right = NULL;
                        }

                        // delete the item now
                        _shmap_btree_item_delete(p_mem, btree, cleanup);
                }

                if ( btree && btree->left != NULL )
                        btree = btree->left;
                if ( btree && btree->right != NULL )
                        btree = btree->right;
        }

        return 1;
}

_shmap_btree * _shmap_btree_item_init ( void * p_mem, _shmap_value * value )
{
        _shmap_btree * ret;

        ret = sh_mem_malloc(p_mem, sizeof(_shmap_btree));

        CHECK_NULL(ret)

        ret->left   = NULL;
        ret->right  = NULL;
        ret->parent = NULL;
        ret->value  = value;

        return ret;
}

int _shmap_btree_item_place ( void * p_mem, _shmap_btree_top * top, _shmap_btree * item, shmap_cleanup cleanup )
{
        int  loadpos, status = 2;
        _shmap_btree * btree = top->tree;

        // calculate, if we need to add it to the left or right side
        // be sute, that we don't generate segfault here
        if ( btree )
        {
                if ( btree->value->keyhash <= item->value->keyhash )
                        loadpos = -1;
                else
                        loadpos = 1;
        }

        while ( btree )
        {
                if ( btree->value->keyhash < item->value->keyhash )
                {
                        // go right, because the value is too small
                        if ( btree->right != NULL )
                                btree = btree->right;
                        else
                                break;
                }

                if ( btree->value->keyhash > item->value->keyhash )
                {
                        // go left, because the value is too big
                        if ( btree->left != NULL )
                                btree = btree->left;
                        else
                                break;
                }

                // if keyhashes are equal, then this is the place to add it
                if ( btree->value->keyhash == item->value->keyhash )
                        break;
        }

        // alright, we should have correct btree position here
/*        if ( btree->parent == NULL && top->tree != NULL )
        {
                printf("Error, error, btree tree currupted??");
                return 0;
        }*/

        if ( btree == NULL ) // we have a toplevel item
        {
                status = 2;
                top->tree = item;
        }
        else // add new item to somewhere in the edge
        {
                if ( btree->value->keyhash < item->value->keyhash )
                {
                        btree->right = item;
                }
                else
                if ( btree->value->keyhash > item->value->keyhash )
                {
                        btree->left = item;
                }
                else
                if ( btree->value->keyhash == item->value->keyhash )
                {
                        // what we need to do, is search for replacement
                        while ( btree->value->keyhash == item->value->keyhash )
                        {
                                if ( !strcmp(btree->value->key, item->value->key) )
                                {
                                        shmap_value_replace(p_mem, btree->value, item->value, NULL);
                                        return 1;
                                }
                        }
                }

        }

        return status;
}

int _shmap_btree_item_delete ( void * p_mem, _shmap_btree * item, shmap_cleanup cleanup )
{
        // check if the item is NULL or not
        CHECK_NULL0(item)

        shmap_value_dispose(p_mem, item->value, cleanup);
        item->value = NULL;
        item->left  = NULL;
        item->right = NULL;
        item->parent= NULL;
        sh_mem_free(p_mem, item);

        return 1;
}
