/***************************************************************************
                          shmap_fifo.c  -  description
                             -------------------
    begin                : Sat Apr 27 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/

#include "shmap_fifo.h"

// this one is the function holder what we use to keep different fifo-s inside
shmap_module * fhold = &ll_module;

shmap_module fifo_module =
        {
        SHMAP_FIFO_MODULE,
        SHMAP_FIFO_VERSION,
        SHMAP_TYPE_FIFO,
        shmap_fifo_init,
        shmap_fifo_get_value,
        shmap_fifo_add_value,
        shmap_fifo_del_value,
        shmap_fifo_get_sub,
        shmap_fifo_del_sub,
        shmap_fifo_clear,
        shmap_fifo_dispose,
        NULL,
        NULL,
        NULL,
        NULL
        };


/**
 *
 *  We will only need to initialize the base(parent shmap), because
 *   all the fifos will be initialized when adding to an empty fifo
 *
 */
void * shmap_fifo_init ( void * p_mem, void * p_opts )
{
        void * ret;

        ret = fhold->shmap_init(p_mem, p_opts);

        return ret;
}

/*
 *  Note, that if you add a deletable object to fifo, then it will be free-d when
 *  shmap_get is done
 */
int shmap_fifo_add_value ( shmap * map, const char * key, void * value, int opts, shmap_cleanup cleanup )
{
        unsigned int keyflag, valflag;
        _shmap_fifo * fifo = NULL;
        _shmap_value * shvalue = NULL;

        shvalue = fhold->shmap_get_value(map, key, 0);

        if ( shvalue )
                fifo = shvalue->value;

        if ( fifo == NULL )
        {
                // if there if no fifo, then we create one
                fifo = _shmap_fifo_init(map->mem_segm);
        }

        if ( !_shmap_fifo_add(fifo, value, opts) )
        {
                return 0;
        }

        // we need to add it to list, so that value type is undeletable
        keyflag = opts & __SHMAP_FREE_KEY;
        valflag = 0xfffffffd;
        // we don't use cleanup function as we are adding only a pointer to list
        return fhold->shmap_add_value(map, key, fifo, keyflag&valflag, NULL);
}


/**
 *
 * IMPORTANT:
 * Now the case here is, that when we return a value, then it uses allocated memory
 * somehow we need to free that resource after we don't need it any more
 * use sh_mem_free when not needed any more
 *
 */
_shmap_value * shmap_fifo_get_value ( shmap * map, const char * key, int cache )
{
        _shmap_value * value = NULL;
        _shmap_fifo  * fifo  = NULL;

        value = fhold->shmap_get_value(map, key, 0);

        if ( value )
                fifo = value->value;

        if ( fifo == NULL )
        {
                printf("no value!'\n");
                return NULL;
        }

        value = _shmap_fifo_get(fifo);

        return value;
}

/**
 *
 *   Deletes all value under a specific key named fifo.
 *
 */
int shmap_fifo_del_value ( shmap * map, const char * key, shmap_cleanup cleanup )
{
        _shmap_value * vholder = fhold->shmap_get_value(map, key, 0);
        _shmap_fifo  * fifo    = (_shmap_fifo *) vholder->value;

        // now delete the entry itself
        #warning kas me ei pea siin mitte mingit pool-i puhastama??
        fhold->shmap_del_value(map, key, NULL);

        // now delete fifo values internally as well
        _shmap_fifo_del(fifo);

        return 1;
}

_shmap_value * shmap_fifo_get_sub ( shmap * map, const char * key, int keylen, int pos, int cached )
{
        _shmap_value * value = NULL;
        _shmap_fifo  * fifo  = NULL;

        value = fhold->shmap_get_sub(map, key, strlen(key), pos, cached);

        if ( value )
                fifo = value->value;

        if ( fifo == NULL )
                return NULL;

        value = _shmap_fifo_get(fifo);

        return value;
}

/**
 *
 *   Deletes all the subvalues from fifo
 *
 */
int shmap_fifo_del_sub ( shmap * map, const char * key, int keylen, shmap_cleanup cleanup )
{
        _shmap_value * fifoholder;
        _shmap_fifo  * fifo;

        while ( (fifoholder = fhold->shmap_get_sub(map, key, keylen, 0, 0)) )
        {
        #warning siin on midagi vist paigast 2ra?
                fifo = (_shmap_fifo *) fifoholder->value;
                fhold->shmap_del_value(map, fifoholder->key, NULL);
                _shmap_fifo_del(fifo);
        }

        return 1;
}

int shmap_fifo_clear ( shmap * map, shmap_cleanup cleanup )
{
        return fhold->shmap_clear(map, _shmap_fifo_cleanup);
}

int shmap_fifo_dispose ( shmap * map, shmap_cleanup cleanup )
{
        return fhold->shmap_clear(map, _shmap_fifo_cleanup);
}

/************************************************************************
 *                     FIFO internals
 *
 ************************************************************************/

/**
 *  free the fifo
 */
void _shmap_fifo_del ( _shmap_fifo * fifo )
{
        // delete all values from fifo
        //   as they are pooled, then we can do it all at once (including the fifo itself)
        sh_pool_dispose(fifo->pool);
}


/**
 *
 *  Cleanup function for values
 *
 */
void _shmap_fifo_cleanup ( _shmap_value * value, int * opts )
{
        _shmap_fifo * fifo = (_shmap_fifo *) value;

        CHECK_NULL_(fifo)

        sh_pool_dispose(fifo->pool);
}

_shmap_value * _shmap_fifo_get ( _shmap_fifo * fifo )
{
        _shmap_value     * ret  = NULL;
        _shmap_fifo_list * free = NULL;
        CHECK_NULL(fifo)

        if ( fifo->toplevel != NULL )
        {
                free = fifo->toplevel;
                fifo->toplevel = fifo->toplevel->next;
                ret = free->value;
                sh_pool_free(fifo->pool, free);
        }

        return ret;
}

_shmap_fifo * _shmap_fifo_init ( void * p_mem )
{
        _shmap_fifo * fifo;
        _shpool * pool;

        pool = sh_pool_init(p_mem);

        CHECK_NULL(pool)

        fifo = sh_pool_malloc(pool, sizeof(_shmap_fifo));

        CHECK_NULL(fifo)

        fifo->pool     = pool;
        fifo->toplevel = NULL;

        return fifo;
}

int _shmap_fifo_add ( _shmap_fifo * fifo, void * value, int opts )
{
        _shmap_fifo_list * item;
        _shmap_fifo_list * browse;

        item = sh_pool_malloc(fifo->pool, sizeof(_shmap_fifo_list));
        CHECK_NULL0(item)
        item->value = shmap_pool_value_init(fifo->pool, "", 0, value, opts);
        CHECK_NULL0(item->value)

        browse = fifo->toplevel;
        if ( browse )
                while ( browse->next != NULL )
                {
                        browse = browse->next;
                }

        if ( browse )
                browse->next = item;
        else
                fifo->toplevel = item;

        return 1;
}

