/***************************************************************************
                          shmap_ll.c  -  description
                             -------------------
    begin                : Tue Feb 5 2002
    copyright            : (C) 2002 by Margus Laak
    email                : margus@mail.ee
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA *
 ***************************************************************************/


/*
        This contains linked list type map functions
*/

#define  __SHMAP_LL_NEED_CORE
#include "shmap_ll.h"

struct _shmap_entry
        {
        _shmap_entry * next;
        _shmap_entry * prev;

        _shmap_value * value;
        };

typedef struct _shmap_ll_cache _shmap_ll_cache;
struct _shmap_ll_cache
        {
        // we sometimes need to use caching for shmap_getsub
        void * cache_key;
        void * cache_toplevel;
        int    cache_keyhash;
        int    cache_pos;
        // cache for shmap_get ... cache is set by shmap_getsub
        void * cache_getkey;
        void * cache_entry;
        };


shmap_module ll_module =
        {
        SHMAP_LL_MODULE,
        SHMAP_LL_VERSION,
        SHMAP_TYPE_LL,
        shmap_ll_init,
        shmap_ll_get_value,
        shmap_ll_add_value,
        shmap_ll_del_value,
        shmap_ll_get_sub,
        shmap_ll_del_sub,
        shmap_ll_clear,
        shmap_ll_dispose,
        shmap_ll_cache_init,
        shmap_ll_cache_clean,
        shmap_ll_cache_dispose,
        shmap_ll_dump
        };

/*
        Initialises linked list map
*/
void * shmap_ll_init ( void * mem_segm, void * p_opts )
{
        return NULL;
}

/*
        function returns value from map

        value can be NULL

        map    : the linked list map itself
        kay    : the key based on what the lookup is done
        cached : 1 if we are allowed to use cache, 0 if we are not allowed to use cache

        NOTES:
                keep in mind, that if the key pointer is the same then value is taken from
                  map cache and no list lookup is done
*/
_shmap_value * shmap_ll_get_value ( shmap * map, const char * key, int cached )
{
        int keyhash;
        _shmap_entry * entry;
        _shmap_ll_cache * cache;

        cache = (_shmap_ll_cache *) map->cache;

        keyhash = _core_get_hashvalue(key);

        if ( cached && cache != NULL && key == cache->cache_getkey && cache->cache_entry != NULL )
                entry = cache->cache_entry;
        else
        {
                entry = _shmap_core_get_entry(map->toplevel, keyhash, key);
                if ( cache != NULL )
                {
                        cache->cache_entry  = entry;
                        cache->cache_getkey = (char *) key;
                }
        }

        if ( entry != NULL )
                return entry->value;
        else
                return NULL;
}

/*
        add new entry to map

        NOTES:
                this function also cleans cache
*/
int shmap_ll_add_value ( shmap * map, const char * key, void * value, int val_opts, shmap_cleanup cleanup )
{
        int status, keyhash;
        _shmap_entry * entry;

        if ( key == NULL || map == NULL )
                return 0;

        keyhash = _core_get_hashvalue(key);

        // first try to replace the value
        status = _shmap_core_replace_value (map->mem_segm, map->toplevel, keyhash, key, value, cleanup);

        if ( status )
                return 1;

        entry = sh_mem_malloc(map->mem_segm, sizeof(_shmap_entry));

        if ( entry == NULL )
                return 0;

        entry->value = shmap_value_init(map->mem_segm, (void *)key, keyhash, value, val_opts);
        if ( entry->value == NULL )
        {
                sh_mem_free(map->mem_segm, entry);
                return 0;
        }

        entry->value->keyhash   = keyhash;

        shmap_ll_cache_clean(map->cache);

        // if core adding fails we take it, that there are no loose links
        status = _shmap_core_add_entry((_shmap_entry **)&map->toplevel, entry);

        // if adding fails, then we don't delete key and value pair
        // we let the caller descide on this one
        if ( !status )
        {
                shmap_value_dispose(map->mem_segm, entry->value, NULL);
                sh_mem_free(map->mem_segm, entry->value);
                sh_mem_free(map->mem_segm, entry);
                return 0;
        }
        else
                return 2;

        return 2;
}

/*
        deletes an entry in the map
*/
int shmap_ll_del_value ( shmap * map, const char * key, shmap_cleanup cleanup )
{
        int keyhash;
        _shmap_entry * entry;
        
        if ( map == NULL )
                return 0;
        
        keyhash = _core_get_hashvalue(key);
        
        shmap_ll_cache_clean(map->cache);

        // get the pointer from list
        entry = _shmap_core_get_entry(map->toplevel, keyhash, key);

        if ( entry == NULL )
                return 0;

        // unlink entry pointer from list
        _shmap_core_del_entry((_shmap_entry **)&map->toplevel, entry);

        // and now dispose the entry itself
        shmap_value_dispose(map->mem_segm, entry->value, cleanup);
        entry->value = NULL;
        sh_mem_free(map->mem_segm, entry);
        entry = NULL;

        return 1;
}

/*
        returns the full key based on the subkey provided

        NOTES:
                this method uses caching. So if you do subsequent queries, then
                  we don't start from the begginning of the list, but where we left off.
                  you can turn caching off by setting cached to 0
*/
_shmap_value * shmap_ll_get_sub ( shmap * map, const char * subkey, int key_len, int pos, int cached )
{
        int keyhash;
        _shmap_entry * entry;
        _shmap_ll_cache * cache;

        if ( map == NULL )
                return NULL;

        cache = map->cache;

        if ( cached && cache != NULL && subkey == cache->cache_key && cache->cache_pos == (pos - 1) )
                entry = _shmap_core_get_subentry(cache->cache_toplevel, cache->cache_keyhash, subkey, key_len, 1);
        else
        {
                keyhash = _core_get_hashvalue(subkey);
                entry   = _shmap_core_get_subentry(map->toplevel, keyhash, subkey, key_len, pos);
                if ( cache != NULL )
                        cache->cache_keyhash  = keyhash;
        }

        // init caching for later use
        if ( cache != NULL )
        {
                cache->cache_key      = (void *) subkey;
                cache->cache_toplevel = entry;
                cache->cache_pos      = pos;

                if ( entry != NULL )
                {
                        cache->cache_getkey   = entry->value->key;
                        cache->cache_entry    = entry;
                }
                else
                {
                        cache->cache_getkey   = NULL;
                        cache->cache_entry    = NULL;
                }
        }

        if ( entry != NULL )
                return entry->value;
        else
                return NULL;
}

int shmap_ll_del_sub ( shmap * map, const char * key, int key_len, shmap_cleanup cleanup )
{
        int keyhash;
        _shmap_entry * entry;

        if ( map == NULL )
                return 0;

        keyhash = _core_get_hashvalue(key);

        shmap_ll_cache_clean(map->cache);

        entry = _shmap_core_get_subentry(map->toplevel, keyhash, key, key_len, 0);

        while ( entry != NULL )
        {
                _shmap_core_del_entry((_shmap_entry **)&map->toplevel, entry);
                shmap_value_dispose(map->mem_segm, entry->value, cleanup);
                entry->value = NULL;
                sh_mem_free(map->mem_segm, entry);
                entry = NULL;
                entry = _shmap_core_get_subentry(map->toplevel, keyhash, key, key_len, 0);
        }
        
        return 1;
}

/*
        deletes only map values. does not destry the whole map
*/
int shmap_ll_clear ( shmap * map, shmap_cleanup cleanup )
{
        int status;

        shmap_ll_cache_clean(map->cache);

        status = _shmap_delv(map, cleanup);
        
        return status;
}

/*
        deletes all map values and also the map itself
        actually the clear and dispose implementation is the same for LL type
*/
int shmap_ll_dispose ( shmap * map, shmap_cleanup cleanup )
{
        return _shmap_delv(map, cleanup);
}


/*
        initializes linked list hashmap caching structure
*/
void * shmap_ll_cache_init ( void * p_mem )
{
        _shmap_ll_cache * cache = sh_mem_malloc(p_mem, sizeof(_shmap_ll_cache));

        shmap_ll_cache_clean(cache);

        return cache;
}

/*
        cleans the internal cache holders
*/
void shmap_ll_cache_clean ( void * p_cache )
{
        _shmap_ll_cache * cache = (_shmap_ll_cache *) p_cache;

        CHECK_NULL_(p_cache)

        cache->cache_entry    = NULL;
        cache->cache_getkey   = NULL;
        cache->cache_key      = NULL;
        cache->cache_keyhash  = 0;
        cache->cache_pos      = 0;
        cache->cache_toplevel = NULL;
}

/*
        distorys the linked list hashmap caching stucture
*/
int shmap_ll_cache_dispose ( void * p_mem, void ** p_cache )
{
        _shmap_ll_cache * cache = (_shmap_ll_cache *) *p_cache;

        sh_mem_free(p_mem, cache);
        *p_cache = NULL;

        return 1;
}

/*
    function dumps the while map to stderr
*/
void shmap_ll_dump ( shmap * map, int with_values )
{
        int i = 0;
        _shmap_entry *entry;
        _shmap_ll_cache * cache;
    
        if ( map == NULL )
                return;
        
        fprintf(stderr, "\n\nMap at: %p\n", map);
        fprintf(stderr, "  toplevel: %p\n", map->toplevel);
        fprintf(stderr, "  mem_segm: %p\n", map->mem_segm);
        fprintf(stderr, "  items:    %d\n", map->items);
        cache = map->cache;
        if ( cache != NULL )
        {
                if ( with_values )
                        fprintf(stderr, "  cache_key:        %p:%s\n", cache->cache_key, (char *) cache->cache_key);
                else
                        fprintf(stderr, "  cache_key:        %p\n", cache->cache_key);
                fprintf(stderr, "  cache_toplevel:   %p\n", cache->cache_toplevel);
                fprintf(stderr, "  cache_keyhash:    %d\n", cache->cache_keyhash);
                fprintf(stderr, "  cache_pos:        %d\n", cache->cache_pos);
                fprintf(stderr, "  cache_getkey:     %p\n", cache->cache_getkey);
                fprintf(stderr, "  cache_entry:      %p\n", cache->cache_entry);
        }

        entry = map->toplevel;

        while ( entry != NULL )
        {
                if ( with_values )
                {
                        fprintf(stderr, "%d) next:%p, prev:%p\n", i, entry->next, entry->prev);
                        fprintf(stderr, "%d) key:%p:%s\n%d value:%p:%s\n", i, entry->value->key, entry->value->key, i, entry->value->value, (const char *) entry->value->value);
                        fprintf(stderr, "%d) keyhash:%d\n", i, entry->value->keyhash);
                }
                else
                {
                        fprintf(stderr, "%d) next:%p, prev:%p\n", i, entry->next, entry->prev);
                        fprintf(stderr, "%d) key:%p\n%d value:%p\n", i, entry->value->key, i, entry->value->value);
                        fprintf(stderr, "%d) keyhash:%d\n", i, entry->value->keyhash);
                }
                entry = entry->next;
                i++;
        }
}


/****************************************************************

        core functions start here
        
        core functions presume, that all hash locks are in place
        core functions mostly deal with pointer associations
        core functions are aware of entry internal life (like
          keys and keyhashes)

*****************************************************************/

static _shmap_entry *_shmap_core_get_entry ( _shmap_entry * toplevel, int keyhash, const char * key )
{
        _shmap_entry * entry = toplevel;

        while ( entry != NULL )
        {
                if ( entry->value->keyhash == keyhash && strcmp(entry->value->key, key) == 0 )
                        return entry;

                entry = entry->next;
        }
        
        return NULL;
}

static int _shmap_core_add_entry ( _shmap_entry **toplevel, _shmap_entry * entry )
{
        _shmap_entry * tmp = *toplevel;

        *toplevel = entry;
        (*toplevel)->next = tmp?tmp:NULL;
        (*toplevel)->prev = NULL;

        if ( tmp != NULL )
                tmp->prev = *toplevel;

        return 1;
}

/*
 * does not free any resources, only relinks the entry from the chains
 * so that it can be free-d later
 */
static int _shmap_core_del_entry ( _shmap_entry **toplevel, _shmap_entry * entry )
{
        if ( entry->next != NULL && entry->prev != NULL ) // item somewhere in middle
        {
                entry->next->prev = entry->prev;
                entry->prev->next = entry->next;
        }
        else
        if ( entry->next == NULL && entry->prev != NULL ) // item in the end
        {
                entry->prev->next = NULL;
        }
        else
        if ( entry->next != NULL && entry->prev == NULL ) // item in the begginning
        {
                entry->next->prev = NULL;
                *toplevel = entry->next;
        }
        else
        if ( entry->next == NULL && entry->prev == NULL ) // the only item in list
        {
                *toplevel = NULL;
        }
        
        return 1;
}

_shmap_entry * _shmap_core_get_subentry ( _shmap_entry *toplevel, int keyhash, const char * key, int key_len, int pos )
{
        int count = 0;
        
        _shmap_entry * entry = toplevel;
        
        while ( entry != NULL )
        {
                if ( strncmp(entry->value->key, key, key_len) == 0 )
                {
                        if ( pos == count )
                                return entry;
                        count++;
                }
                
                entry = entry->next;
        }
        
        return NULL;
}

static int _shmap_delv ( shmap * map, shmap_cleanup cleanup )
{
        _shmap_entry * entry = map->toplevel, * free;


        while ( entry != NULL )
        {
                free = entry;
                entry = entry->next;

                shmap_value_dispose(map->mem_segm, free->value, cleanup);

                free->next = NULL;
                free->prev = NULL;
                free->value = NULL;

                if ( free != NULL )
                        sh_mem_free(map->mem_segm, free);
                free       = NULL;
        }

        map->toplevel = NULL;
        
        return 1;
}

/*
        only does a replace in list. if key doesn't exist, then replace is not done

        NOTES: value will not be copyed, only the pointer value will be replaced
                (so you'll have to supply already strdup-d value)
*/
static int _shmap_core_replace_value ( void * p_mem, _shmap_entry * toplevel, int keyhash, const char * key, const char * value, shmap_cleanup cleanup )
{
       _shmap_entry * entry = toplevel;
       
       while ( entry != NULL )
       {
                if ( entry->value->keyhash == keyhash && strcmp(entry->value->key, key) == 0 )
                {
                        shmap_value_replace_value(p_mem, entry->value, (void *)value, cleanup);
                        return 1;
                }
                entry = entry->next;
       }

       return 0;
}

